﻿using System.Net;
using System.Runtime.InteropServices;
using System.Text;
using System.Text.RegularExpressions;
using Microsoft.Extensions.Logging;
using Newtonsoft.Json;
using Newtonsoft.Json.Linq;

namespace SteamGameTimeTrack;

public class API : IDisposable
{
    private readonly HttpListener _server = new();
    private Thread listenThread;
    private CancellationTokenSource cts = new();
    private ILogger logger;
    private readonly Tracker parent;

    public API(Tracker parent, int port, ILogger logger)
    {
        this.parent = parent;
        this.logger = logger;
        if(RuntimeInformation.IsOSPlatform(OSPlatform.Linux))
            this._server.Prefixes.Add($"http://*:{port}/");
        else
            this._server.Prefixes.Add($"http://localhost:{port}/");
        this.listenThread = new (Listen);
        this.listenThread.Start();
    }
    
    private void Listen()
    {
        this._server.Start();
        foreach (string serverPrefix in this._server.Prefixes)
            logger.LogInformation($"Listening on {serverPrefix}");
        while (this._server.IsListening)
        {
            try
            {
                //cts.CancelAfter(1000);
                HttpListenerContext context = this._server.GetContext();
                //Log($"{context.Request.HttpMethod} {context.Request.Url} {context.Request.UserAgent}");
                Task t = new(() =>
                {
                    HandleRequest(context);
                });
                t.Start();
            }
            catch (Exception e)
            {
                this.logger.LogError(e, e.ToString());
            }
        }
    }
    
    private void HandleRequest(HttpListenerContext context)
    {
        HttpListenerRequest request = context.Request;
        HttpListenerResponse response = context.Response;
        if (request.Url!.LocalPath.Contains("favicon"))
        {
            SendResponse(HttpStatusCode.NoContent, response);
            return;
        }

        switch (request.HttpMethod)
        {
            case "GET":
                HandleGet(request, response);
                break;
            case "OPTIONS":
                SendResponse(HttpStatusCode.OK, response);
                break;
            default: 
                SendResponse(HttpStatusCode.MethodNotAllowed, response);
                break;
        }
    }
    
    private Dictionary<string, string> GetRequestVariables(string query)
    {
        Dictionary<string, string> ret = new();
        Regex queryRex = new (@"\?{1}&?([A-z0-9-=]+=[A-z0-9-=]+)+(&[A-z0-9-=]+=[A-z0-9-=]+)*");
        if (!queryRex.IsMatch(query))
            return ret;
        query = query.Substring(1);
        foreach (string keyValuePair in query.Split('&').Where(str => str.Length >= 3))
        {
            string var = keyValuePair.Split('=')[0];
            string val = Regex.Replace(keyValuePair.Substring(var.Length + 1), "%20", " ");
            val = Regex.Replace(val, "%[0-9]{2}", "");
            ret.Add(var, val);
        }
        return ret;
    }


    private void HandleGet(HttpListenerRequest request, HttpListenerResponse response)
    {
        Dictionary<string, string> requestVariables = GetRequestVariables(request.Url!.Query);
        string path = Regex.Match(request.Url!.LocalPath, @"[A-z0-9]+(\/[A-z0-9]+)*").Value;
        Regex playerInfoRex = new(@"Player(?:\/([0-9]+))?");
        Regex sessionInfoRex = new(@"Player\/([0-9]+|[A-z0-9]+)\/Session(?:\/([0-9]+))?");
        Regex gameTimeInfoRex = new(@"Player\/([0-9]+|[A-z0-9]+)\/GameTime(?:\/([0-9]+))?");
        Regex playerGameInfo = new(@"Player\/([0-9]+|[A-z0-9]+)(?:\/([0-9]+))");
        
        this.logger.LogInformation($"API GET {path}");

        if (gameTimeInfoRex.Match(path) is { Success: true } gti && gti.Value.Equals(path))
        {
            if (parent._players.Any(p => p.steamid.Equals(gti.Groups[1].Value)))
            {
                Player player = parent._players.First(p => p.steamid.Equals(gti.Groups[1].Value));
                if (gti.Groups[2].Success && int.TryParse(gti.Groups[2].Value, out int appid)) //Return specific game
                {
                    if (player.GameTimes.Any(g => g.appid == appid))
                        SendResponse(HttpStatusCode.OK, response, player.GameTimes.First(g => g.appid == appid));
                    else
                        SendResponse(HttpStatusCode.NotFound, response, $"AppId {appid} not found");
                }
                else //Return all games
                {
                    SendResponse(HttpStatusCode.OK, response, player.GameTimes);
                }
            }
            else
                SendResponse(HttpStatusCode.NotFound, response, $"Player {gti.Groups[1].Value} not found");
        }else if (sessionInfoRex.Match(path) is { Success: true } si&& si.Value.Equals(path))
        {
            if (parent._players.Any(p => p.steamid.Equals(si.Groups[1].Value)))
            {
                Player player = parent._players.First(p => p.steamid.Equals(si.Groups[1].Value));
                if (si.Groups[2].Success && int.TryParse(si.Groups[2].Value, out int appid)) //Return specific game
                {
                    if (player.Sessions.Any(s => s.AppId == appid))
                        SendResponse(HttpStatusCode.OK, response, player.Sessions.Where(s => s.AppId == appid));
                    else
                        SendResponse(HttpStatusCode.NotFound, response, $"No Sessions for AppId {appid} found");
                }
                else //Return all games
                {
                    SendResponse(HttpStatusCode.OK, response, player.Sessions);
                }
            }
            else
                SendResponse(HttpStatusCode.NotFound, response, $"Player {si.Groups[1].Value} not found");
        }else if(playerGameInfo.Match(path) is { Success: true } gi&& gi.Value.Equals(path))
        {
            if (parent._players.Any(p => p.steamid.Equals(gi.Groups[1].Value)))
            {
                Player player = parent._players.First(p => p.steamid.Equals(gi.Groups[1].Value));
                if (gi.Groups[2].Success && int.TryParse(gi.Groups[2].Value, out int appid)) //Return specific game
                {
                    JObject x = new();
                    if (player.Sessions.Any(s => s.AppId == appid))
                        x["Sessions"] = new JArray(player.Sessions.Where(s => s.AppId == appid));
                    else
                        x["Sessions"] = null;
                    if (player.GameTimes.Any(g => g.appid == appid))
                        x["GameTime"] = JObject.FromObject(player.GameTimes.First(g => g.appid == appid));
                    else
                        x["GameTime"] = null;
                    
                    if(x["Sessions"] != null && x["GameTime"] != null)
                        SendResponse(HttpStatusCode.OK, response, x);
                    else
                        SendResponse(HttpStatusCode.NotFound, response, $"AppId {appid} not found");
                }
                else //Return all games
                {
                    SendResponse(HttpStatusCode.BadRequest, response);
                }
            }
            else
                SendResponse(HttpStatusCode.NotFound, response, $"Player {gi.Groups[1].Value} not found");
        }else if (playerInfoRex.Match(path) is { Success: true } pi&& pi.Value.Equals(path))
        {
            if (pi.Groups[1].Success)//Return specific Player
            {
                if(parent._players.Any(p => p.steamid.Equals(pi.Groups[1].Value)))
                    SendResponse(HttpStatusCode.OK, response, parent._players.First(p => p.steamid.Equals(pi.Groups[1].Value)));
                else
                    SendResponse(HttpStatusCode.NotFound, response, $"Player {pi.Groups[1].Value} not found");
            }
            else//Return all Players
                SendResponse(HttpStatusCode.OK, response, parent._players);
        }else
            SendResponse(HttpStatusCode.NotFound, response, "Path not found.\n" +
                                                            "Valid Paths:\n" +
                                                            "\t/Player/<SteamID64>\n" +
                                                            "\t/Player/<SteamID64>/<appid>\n" +
                                                            "\t/Player/<SteamID64>/GameTime/<appid>\n" +
                                                            "\t/Player/<SteamID64>/Session/<appid>\n");
    }

    private void SendResponse(HttpStatusCode statusCode, HttpListenerResponse response, object? content = null)
    {
        logger.LogInformation($"Response: {statusCode.ToString()}");
        response.StatusCode = (int)statusCode;
        response.AddHeader("Access-Control-Allow-Headers", "Content-Type, Accept, X-Requested-With");
        response.AddHeader("Access-Control-Allow-Methods", "GET");
        response.AddHeader("Access-Control-Max-Age", "5");
        response.AppendHeader("Access-Control-Allow-Origin", "*");

        byte[] bytes;
        if (content is not null)
        {
            if (content is string str)
                bytes = Encoding.UTF8.GetBytes(str);
            else if (content is JObject jo)
                bytes = Encoding.UTF8.GetBytes(jo.ToString());
            else
                bytes = Encoding.UTF8.GetBytes(JsonConvert.SerializeObject(content));
        }
        else
            bytes = Array.Empty<byte>();

        if (content is not Stream)
        {
            response.ContentType = content?.GetType() == typeof(string) ? "text/plain" : "application/json";
            try
            {
                response.OutputStream.Write(bytes);
                response.OutputStream.Close();
            }
            catch (HttpListenerException e)
            {
                logger.LogError(e, null);
            }
        }
        else if(content is FileStream stream)
        {
            string contentType = stream.Name.Split('.')[^1];
            switch (contentType.ToLower())
            {
                case "gif":
                    response.ContentType = "image/gif";
                    break;
                case "png":
                    response.ContentType = "image/png";
                    break;
                case "jpg":
                case "jpeg":
                    response.ContentType = "image/jpeg";
                    break;
                case "log":
                    response.ContentType = "text/plain";
                    break;
            }
            stream.CopyTo(response.OutputStream);
            response.OutputStream.Close();
            stream.Close();
        }
    }

    public void Dispose()
    {
        ((IDisposable)_server).Dispose();
    }
}