﻿// ReSharper disable IdentifierTypo
// ReSharper disable InconsistentNaming
// ReSharper disable MemberCanBePrivate.Global
// ReSharper disable UnassignedField.Global
// ReSharper disable PropertyCanBeMadeInitOnly.Global

using System.Reflection;
using Newtonsoft.Json;
using Newtonsoft.Json.Linq;

namespace SteamGameTimeTrack;

public struct Player
{
    private readonly Net net;
    
    public string steamid { get; init; }
    public bool IsMe { get; init; }

    public PersonaStateEnum personastate { get; set; } = PersonaStateEnum.Offline;
    public byte communityvisibilitystate { get; set; } = 0;
    public bool profilestate { get; set; } = false;
    public long lastlogoff { get; set; } = 0;
    public bool? commentpermission { get; set; } = null;

    public string? realname { get; set; } = null;
    public string? primaryclanid { get; set; } = null;
    public string? gameserverip { get; set; } = null;
    public string? gameextrainfo { get; set; } = null;
    public string? loccountrycode { get; set; } = null;
    public string? locstatecode { get; set; } = null;
    public string? loccityid { get; set; } = null;
    public string? personaname { get; set; } = null;
    public string? profileurl { get; set; } = null;
    public string? avatar { get; set; } = null;
    public string? avatarmedium { get; set; } = null;
    public string? avatarfull { get; set; } = null;
    public string? gameid { get; set; } = null;
    public long timecreated { get; set; } = 0;
    public long friend_since { get; set; } = 0;
    public HashSet<GameTime> GameTimes = new();
    // ReSharper disable once CollectionNeverQueried.Global
    public List<Session> Sessions = new();

    [JsonIgnore]
    public DateTime? TimeCreated => DateTime.UnixEpoch.AddSeconds(timecreated);
    [JsonIgnore]
    public DateTime? FriendSince => DateTime.UnixEpoch.AddSeconds(friend_since);
    [JsonIgnore]
    public DateTime? LastLogoff => DateTime.UnixEpoch.AddSeconds(lastlogoff);

    [JsonConstructor]
    public Player(string steamid, Net net, bool IsMe = false)
    {
        this.IsMe = IsMe;
        this.net = net;
        this.steamid = steamid;
    }

    public override int GetHashCode()
    {
        return steamid.GetHashCode();
    }

    public void Export()
    {
        string path = Path.Join("states", steamid);
        if (!Directory.Exists(path))
            Directory.CreateDirectory(path);
        File.WriteAllText(Path.Join(path, $"{DateTime.UtcNow:dd-mm-yy hh-mm-ss}.json"), JsonConvert.SerializeObject(this, Formatting.Indented));
    }

    public bool Diffs(Player other, out string diffStr)
    {
        List<string> diffs = new();
        foreach (FieldInfo field in GetType().GetFields())
        {
            object? thisValue = field.GetValue(this);
            object? otherValue = field.GetValue(other);
            if(thisValue is not null && otherValue is not null && !thisValue.Equals(otherValue))
                diffs.Add($"{field.Name} {thisValue} -> {otherValue}");
        }
        
        foreach (PropertyInfo property in GetType().GetProperties())
        {
            object? thisValue = property.GetValue(this);
            object? otherValue = property.GetValue(other);
            if(thisValue is not null && otherValue is not null && !thisValue.Equals(otherValue))
                diffs.Add($"{property.Name} {thisValue} -> {otherValue}");
        }

        GameTime[] newGameTimes = other.GameTimes.Except(GameTimes).ToArray();
        foreach (GameTime newGameTime in newGameTimes)
            diffs.Add($"New game {newGameTime.name} {newGameTime.appid} {newGameTime.PlayTime}");
        foreach (GameTime otherGameTime in other.GameTimes)
            // ReSharper disable once PatternAlwaysMatches
            if(this.GameTimes.First(tgt => tgt.appid == otherGameTime.appid) is GameTime thisGameTime && !thisGameTime.GameStarted.Equals(otherGameTime.GameStarted))
                diffs.Add($"Started {otherGameTime.name} {otherGameTime.appid} at {otherGameTime.GameStarted}");

        diffStr = !diffs.Any() ? "No diffs" : $"\n\t{string.Join("\n\t", diffs)}";
        return diffs.Any();
    }

    public void UpdateGameInfo(GameTime gameTime)
    {
        if (this.GameTimes.Any(gt => gt.appid == gameTime.appid))
        {
            GameTime thisGameTime = this.GameTimes.First(gt => gt.appid == gameTime.appid);
            this.GameTimes.Remove(thisGameTime);
        }

        this.GameTimes.Add(gameTime);
    }

    internal void GetGames()
    {
        JObject? gamesResult = net.MakeRequestGetJObject($"https://api.steampowered.com/IPlayerService/GetOwnedGames/v0001/?&steamid={steamid}&include_appinfo=true&include_played_free_games=true");
        if (gamesResult is not null && gamesResult.TryGetValue("response", out JToken? value) && value.SelectToken("games") is not null)
        {
            foreach (JToken game in value["games"]!)
            {
                GameTime gameTime = game.ToObject<GameTime>();
                this.UpdateGameInfo(gameTime);
            }
        }
    }

    public Player WithInfo(Player other)
    {
        if (other.gameid is not null && this.gameid is null) //Game started
        {
            if(!this.GameTimes.Any(game => game.appid.ToString().Equals(other.gameid)))
                GetGames();
            GameTime gameTime = this.GameTimes.First(game => game.appid.ToString().Equals(other.gameid));
            this.GameTimes.Remove(gameTime);
            gameTime.GameStarted = DateTime.UtcNow;
            this.GameTimes.Add(gameTime);
        }else if (other.gameid is null && this.gameid is not null) //Game stopped
        {
            if(!this.GameTimes.Any(game => game.appid.ToString().Equals(other.gameid)))
                GetGames();
            GameTime gameTime = this.GameTimes.First(game => game.appid.ToString().Equals(other.gameid));
            this.Sessions.Add(Session.FromGameTime(gameTime));
            this.GameTimes.Remove(gameTime);
            gameTime.GameStarted = null;
            this.GameTimes.Add(gameTime);
        }
        
        return this with
        {
            personastate = other.personastate,
            communityvisibilitystate = other.communityvisibilitystate,
            profilestate = other.profilestate,
            lastlogoff = other.lastlogoff,
            commentpermission = other.commentpermission,
            realname = other.realname,
            primaryclanid = other.primaryclanid,
            gameserverip = other.gameserverip,
            gameextrainfo = other.gameextrainfo,
            loccountrycode = other.loccountrycode,
            locstatecode = other.locstatecode,
            loccityid = other.loccityid,
            personaname = other.personaname,
            profileurl = other.profileurl,
            avatar = other.avatar,
            avatarmedium = other.avatarmedium,
            avatarfull = other.avatarfull,
            gameid = other.gameid
        };
    }

    internal void ShuttingDownSaveSessions()
    {
        if (this.gameid is null)
            return;
        string gid = this.gameid;
        GameTime? gameTime = this.GameTimes.FirstOrDefault(game => game.appid.ToString().Equals(gid));
        if (gameTime is null || gameTime.Value.GameStarted is null)
            return;
        this.Sessions.Add(Session.FromGameTime(gameTime.Value));
        this.gameid = null;
        this.Export();
    }
    
    internal class PlayerJsonConverter : JsonConverter
    {
        private Net net;

        public PlayerJsonConverter(Net net)
        {
            this.net = net;
        }
        
        public override bool CanConvert(Type objectType)
        {
            return objectType == typeof(Player);
        }

        public override object? ReadJson(JsonReader reader, Type objectType, object? existingValue, JsonSerializer serializer)
        {
            JObject jo = JObject.Load(reader);
            Player tmp = new (jo.Value<string>("steamid")!, net, jo.Value<bool>("IsMe"))
            {
                GameTimes = jo["GameTimes"]!.ToObject<HashSet<GameTime>>()!,
                Sessions = jo["Sessions"]!.ToObject<List<Session>>()!
            };
            return tmp.WithInfo(jo.ToObject<Player>());
        }

        public override bool CanWrite => false;
        
        public override void WriteJson(JsonWriter writer, object? value, JsonSerializer serializer)
        {
            throw new NotImplementedException();
        }
    }
}

public enum PersonaStateEnum : byte
{
    Offline = 0,
    Online = 1,
    Busy = 2,
    Away = 3,
    Snooze = 4,
    LookingToTrade = 5,
    LookingToPlay = 6
}